@@ -10,7 +10,7 @@ use Mojo::URL;
 use Scalar::Util 'weaken';
 use SQL::Abstract::mysql;
 
-our $VERSION = '1.21';
+our $VERSION = '1.22';
 
 has abstract        => sub { SQL::Abstract::mysql->new(quote_char => chr(96), name_sep => '.') };
 has auto_migrate    => 0;
@@ -211,10 +211,16 @@ Mojo::mysql - Mojolicious and Async MySQL/MariaDB
     ->hashes->map(sub { $_->{name} })->join("\n")->say;
 
   # Select all rows non-blocking
-  $db->query('select * from names' => sub {
-    my ($db, $err, $results) = @_;
-    $results->hashes->map(sub { $_->{name} })->join("\n")->say;
-  });
+  Mojo::IOLoop->delay(
+    sub {
+      my $delay = shift;
+      $db->query('select * from names' => $delay->begin);
+    },
+    sub {
+      my ($delay, $err, $results) = @_;
+      $results->hashes->map(sub { $_->{name} })->join("\n")->say;
+    }
+  )->wait;
 
   # Concurrent non-blocking queries (synchronized with promises)
   my $now   = $db->query_p('select now() as now');
